from flask import Blueprint
from flask import render_template
from flask import redirect
from flask import flash
from flask import url_for
from flask_login import current_user
from flask_login import login_user
from flask import current_app
from flask import g
from flask_wtf import Form
from wtforms import TextField
from wtforms import PasswordField
from wtforms.validators import InputRequired
from datetime import datetime
from werkzeug.local import LocalProxy
import os
import json
import binascii
import hashlib
from functools import wraps



class CreateUserForm(Form):
    name = TextField('', [InputRequired()])
    password = PasswordField('', [InputRequired()])

def get_users():
    users = getattr(g, '_users', None)
    if users is None:
        users = g._users = UserManager(current_app.config['USER_DIR'])
    return users

bp = Blueprint('wiki', __name__)
current_users = LocalProxy(get_users)


@bp.route('/user/create/', methods=['GET', 'POST'])
def user_create():
    form = CreateUserForm()
    manager = UserManager('user')
    if form.validate_on_submit():
        localtime = str(datetime.now())
        manager.add_user(form.name.data, form.password.data, localtime, True, [], None)
        user = current_users.get_user(form.name.data)
        login_user(user)
        user.set('authenticated', True)
        flash('User successfully created', 'success')
        return redirect(url_for('wiki.index'))
    return render_template('createuser.html', form=form)


class UserManager(object):
    """A very simple user Manager, that saves it's data as json."""
    def __init__(self, path):
        self.file = os.path.join(path, 'users.json')

    def read(self):
        if not os.path.exists(self.file):
            return {}
        with open(self.file) as f:
            data = json.loads(f.read())
        return data

    def write(self, data):
        with open(self.file, 'w') as f:
            f.write(json.dumps(data, indent=2))

    def add_user(self, name, password, time, active=True, roles=[], authentication_method=None):
        users = self.read()

        if users.get(name):
            return False
        if authentication_method is None:
            authentication_method = get_default_authentication_method()
        new_user = {
            'active': active,
            'roles': roles,
            'time': time,
            'authentication_method': authentication_method,
            'authenticated': False
        }

        # Currently we have only two authentication_methods: cleartext and
        # hash. If we get more authentication_methods, we will need to go to a
        # strategy object pattern that operates on User.data.
        if authentication_method == 'hash':
            new_user['hash'] = make_salted_hash(password)
        elif authentication_method == 'cleartext':
            new_user['password'] = password
        else:
            raise NotImplementedError(authentication_method)
        users[name] = new_user
        self.write(users)
        userdata = users.get(name)
        return User(self, name, userdata)

    def getUsers(self):
        users = self.read()
        userarray = []

        for i in users:
            user = User(self,i,users.get(i))
            user.time = users.get(i).get("time")
            userarray.append(user)

        return userarray

    def get_user(self, name):
        users = self.read()
        userdata = users.get(name)
        if not userdata:
            return None
        return User(self, name, userdata)

    def delete_user(self, name):
        users = self.read()
        if not users.pop(name, False):
            return False
        self.write(users)
        return True

    def update(self, name, userdata):
        data = self.read()
        data[name] = userdata
        self.write(data)


class User(object):
    def __init__(self, manager, name, data):
        self.manager = manager
        self.name = name
        self.data = data

    def get(self, option):
        return self.data.get(option)

    def set(self, option, value):
        self.data[option] = value
        self.save()

    def save(self):
        self.manager.update(self.name, self.data)

    def is_authenticated(self):
        return self.data.get('authenticated')

    def is_active(self):
        return self.data.get('active')

    def is_anonymous(self):
        return False

    def get_id(self):
        return self.name

    @property
    def time(self):
        try:
            return self['time']
        except KeyError:
            return self.name

    def time(self, value):
        self['time'] = value

    def check_password(self, password):
        """Return True, return False, or raise NotImplementedError if the
        authentication_method is missing or unknown."""
        authentication_method = self.data.get('authentication_method', None)
        if authentication_method is None:
            authentication_method = get_default_authentication_method()
        # See comment in UserManager.add_user about authentication_method.
        if authentication_method == 'hash':
            result = check_hashed_password(password, self.get('hash'))
        elif authentication_method == 'cleartext':
            result = (self.get('password') == password)
        else:
            raise NotImplementedError(authentication_method)
        return result


def get_default_authentication_method():
    return current_app.config.get('DEFAULT_AUTHENTICATION_METHOD', 'cleartext')


def make_salted_hash(password, salt=None):
    if not salt:
        salt = os.urandom(64)
    d = hashlib.sha512()
    d.update(salt[:32])
    d.update(password)
    d.update(salt[32:])
    return binascii.hexlify(salt) + d.hexdigest()


def check_hashed_password(password, salted_hash):
    salt = binascii.unhexlify(salted_hash[:128])
    return make_salted_hash(password, salt) == salted_hash


def protect(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if current_app.config.get('PRIVATE') and not current_user.is_authenticated:
            return current_app.login_manager.unauthorized()
        return f(*args, **kwargs)
    return wrapper
